#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <string.h>

#include <xen/io/console.h>
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

struct termios  saved_attributes;
int             saved_fl;
int             internal_console = 0;

static void tty_save(void)
{
    fcntl(0,F_GETFL,&saved_fl);
    tcgetattr (0, &saved_attributes);
}

static void tty_raw(void)
{
    struct termios tattr;
    
    fcntl(0,F_SETFL,O_NONBLOCK);
    memcpy(&tattr,&saved_attributes,sizeof(struct termios));
//    tattr.c_lflag &= ~(ISIG|ICANON|ECHO);
    tattr.c_lflag &= ~(ICANON|ECHO);
    tattr.c_cc[VMIN] = 1;
    tattr.c_cc[VTIME] = 0;
    tcsetattr (0, TCSAFLUSH, &tattr);
}

static void tty_restore(void)
{
    fcntl(0,F_SETFL,saved_fl);
    tcsetattr (0, TCSANOW, &saved_attributes);
}

/* ------------------------------------------------------------------ */

static int console_recv(struct xenvm *xen)
{
    uint64_t paddr = frame_to_addr(xen->console_pfn);
    struct xencons_interface *console = guest_paddr_to_ptr(xen, paddr);
    XENCONS_RING_IDX cons, prod, i;
    char buf[sizeof(console->out)];

    cons = console->out_cons;
    prod = console->out_prod;
    /* FIXME: mb */

    if (cons == prod || (prod - cons) > sizeof(console->out))
	return 0;

    for (i = 0; cons != prod; i++, cons++)
	buf[i] = console->out[MASK_XENCONS_IDX(cons, console->out)];

    /* FIXME: mb */
    console->out_cons = cons;

    condev_print(xen, buf, i);
    return 0;
}

static int console_send(struct xenvm *xen, uint8_t *buf, int len)
{
    uint64_t paddr = frame_to_addr(xen->console_pfn);
    struct xencons_interface *console = guest_paddr_to_ptr(xen, paddr);
    XENCONS_RING_IDX cons, prod, i;

    cons = console->in_cons;
    prod = console->in_prod;

    if (prod - cons > sizeof(console->in))
	return -EIO;

    if (sizeof(console->in) - (prod - cons) -1 < len)
	return -EAGAIN;

    /* FIXME: mb */

    for (i = 0; i < len; i++, prod++)
	console->in[MASK_XENCONS_IDX(prod, console->in)] = buf[i];

    /* FIXME: wmb */

    console->in_prod += len;

    raise_event(xen, 0, xen->console_event);
    return 0;
}

static int console_event(struct xenvm *xen, int port, void *data)
{
    if (internal_console || !xen->xenstore) {
	console_recv(xen);
    } else {
	evtchn_notify(xen, xen->console_event);
    }
    return 0;
}

static int console_data(struct xenvm *xen, int fd, void *data)
{
    int tty = 0 /* stdin */;
    static uint8_t buf[16];
    int rc;

    for (;;) {
	rc = read(tty, buf, sizeof(buf));
	switch (rc) {
	case -1:
	    if (EAGAIN == errno)
		return 0;
	    perror("read tty");
	    return -1;
	case 0:
	    fprintf(stderr, "EOF tty\n");
	    return -1;
	default:
	    console_send(xen, buf, rc);
	    break;
	}
    }
}

int console_init(struct xenvm *xen)
{
    int tty = 0 /* stdin */;

    tty_save();
    tty_raw();

    xen->console_event = evtchn_port(xen, "console");
    xenner_register_cb(xen, XENNER_CB_EVTCHN, "console", xen->console_event,
		       NULL, console_event);
    if (internal_console || !xen->xenstore)
	xenner_register_cb(xen, XENNER_CB_IOFD, "console", tty, NULL, console_data);
    return 0;
}

int console_fini(struct xenvm *xen)
{
    tty_restore();
    return 0;
}
