#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>

/* ----------------------------------------------------------- */

static int libxc_trace;
static int libxc_fixme;

#define xc_evtchn_open              xenner_evtchn_open
#define xc_evtchn_fd                xenner_evtchn_fd
#define xc_evtchn_bind_unbound_port xenner_evtchn_bind_unbound_port
#define xc_evtchn_bind_interdomain  xenner_evtchn_bind_interdomain
#define xc_evtchn_bind_virq         xenner_evtchn_bind_virq
#define xc_evtchn_unbind            xenner_evtchn_unbind
#define xc_evtchn_notify            xenner_evtchn_notify
#define xc_evtchn_pending           xenner_evtchn_pending
#define xc_evtchn_unmask            xenner_evtchn_unmask
#define xc_evtchn_close             xenner_evtchn_close

#include <xenctrl.h>
#include "evtchnd.h"
#include "libxc-evtchn.c"

/* ----------------------------------------------------------- */

#include "xenner.h"

int evtchn_port(struct xenvm *xen, char *name)
{
    int port;
    
    if (-1 == xen->evtchnd) {
	/* fake ports */
	xen->next_event++;
	port = xen->next_event;
    } else {
	port = xc_evtchn_bind_unbound_port(xen->evtchnd, 0);
    }
    d1printf("%s: port %d is %s\n", __FUNCTION__, port, name);
    return port;
}

int evtchn_notify(struct xenvm *xen, int port)
{
    if (-1 == xen->evtchnd)
	return 0;
    return xc_evtchn_notify(xen->evtchnd, port);
}

static int evtchn_data(struct xenvm *xen, int fd, void *data)
{
    static int errors = 0;
    int count = 0;
    int port;
    
    if (-1 == xen->evtchnd)
	return 0;

    for (;;) {
	if (!has_data(xen->evtchnd, 0))
	    return count;
	port = xc_evtchn_pending(xen->evtchnd);
	if (port >= 0) {
	    if (port == xen->console_event)
		d3printf("%s: %d pending (console)\n", __FUNCTION__, port);
	    else if (port == xen->xenstore_event)
		d3printf("%s: %d pending (xenstore)\n", __FUNCTION__, port);
	    else
		d2printf("%s: %d pending\n", __FUNCTION__, port);
	    raise_event(xen, 0, port);
	    errors = 0;
	    count++;
	} else if (++errors > 3) {
	    d0printf("%s: repeated evtchn failure, closing\n", __FUNCTION__);
	    evtchn_fini(xen);
	    return count;
	}
    }
    return count;
}

int evtchn_init(struct xenvm *xen)
{
    if (xen->debug)
	libxc_fixme = 1;
    if (xen->debug > 1)
	libxc_trace = 1;
    
    xen->evtchnd = xenner_evtchn_open();
    if (-1 == xen->evtchnd) {
	d0printf("%s: can't connect to evtchnd (not running?)\n", __FUNCTION__);
	return -1;
    }

    xenner_register_cb(xen, XENNER_CB_IOFD, "evtchn", xen->evtchnd, NULL, evtchn_data);
    xenner_evtchnd_domid(xen->evtchnd, xen->domid);
    return 0;
}

void evtchn_fini(struct xenvm *xen)
{
    if (-1 != xen->evtchnd)
	xenner_evtchn_close(xen->evtchnd);
    xen->evtchnd = -1;
}
